from __future__ import annotations

from typing import Any

from sqlalchemy.exc import DatabaseError

import ckan.plugins.toolkit as tk
from ckan.logic import validate
from ckan.types import Context

from ckanext.bulk.entity_manager import get_entity_manager

from . import schema


@validate(schema.bulk_update_entity)
def bulk_update_entity(context: Context, data_dict: dict[str, Any]):
    tk.check_access("bulk_manager", context, data_dict)

    entity_manager = get_entity_manager(data_dict["entity_type"])

    error = None
    result = None

    if data_dict["action"] == "update":
        try:
            result = entity_manager.update_entity(
                data_dict["entity_id"], data_dict["update_on"]
            )
        except tk.ValidationError as e:
            error = str(e.error_dict)
    elif data_dict["action"] == "delete":
        try:
            result = entity_manager.delete_entity(data_dict["entity_id"])
        except tk.ValidationError as e:
            error = str(e.error_dict)
    else:
        error = "Action is not supported"

    response = {
        "result": result,
        "error": error,
        "action": data_dict["action"],
        "entity_id": data_dict["entity_id"],
    }

    if error:
        response["error"] = error

    return response


@validate(schema.bulk_get_entities_by_filters)
def bulk_get_entities_by_filters(context: Context, data_dict: dict[str, Any]):
    tk.check_access("bulk_manager", context, data_dict)

    entity_manager = get_entity_manager(data_dict["entity_type"])

    try:
        result = entity_manager.search_entities_by_filters(
            data_dict["filters"], data_dict["global_operator"]
        )
    except (ValueError, tk.ValidationError) as e:
        return {
            "entities": [],
            "error": str(e),
        }
    except DatabaseError as e:
        return {
            "entities": [],
            "error": f"Database error: {e.statement}",
        }

    return {"entities": result}


@tk.side_effect_free
@validate(schema.bulk_search_fields)
def bulk_search_fields(context: Context, data_dict: dict[str, Any]):
    tk.check_access("bulk_manager", context, data_dict)

    entity_manager = get_entity_manager(data_dict["entity_type"])

    try:
        result = entity_manager.get_fields()
    except tk.ValidationError as e:
        return {
            "result": [],
            "error": str(e),
        }

    return {"result": result}



from __future__ import annotations

import ckan.plugins.toolkit as tk
from ckan import types
from ckan.logic.schema import validator_args

from ckanext.bulk import const
from ckanext.bulk.entity_manager import get_entity_managers


@validator_args
def bulk_get_entities_by_filters(
    not_empty: types.Validator,
    unicode_safe: types.Validator,
    one_of: types.Validator,
    default: types.Validator,
) -> types.Schema:
    entity_types = [v.entity_type for v in get_entity_managers().values()]
    actions = [opt["value"] for opt in tk.h.bulk_action_options()]
    operators = [opt["value"] for opt in tk.h.bulk_operator_options()]

    return {
        "entity_type": [not_empty, unicode_safe, one_of(entity_types)],  # type: ignore
        "action": [not_empty, unicode_safe, one_of(actions)],  # type: ignore
        "filters": {
            "field": [not_empty, unicode_safe],
            "operator": [not_empty, unicode_safe, one_of(operators)],  # type: ignore
            "value": [default(""), unicode_safe],  # type: ignore
        },
        "global_operator": [
            not_empty,
            unicode_safe,
            one_of(
                [
                    const.GLOBAL_AND,
                    const.GLOBAL_OR,
                ]
            ),  # type: ignore
        ],
    }


@validator_args
def bulk_search_fields(
    not_empty: types.Validator,
    unicode_safe: types.Validator,
    one_of: types.Validator,
    default: types.Validator,
) -> types.Schema:
    entity_types = [v.entity_type for v in get_entity_managers().values()]

    return {
        "entity_type": [not_empty, unicode_safe, one_of(entity_types)],  # type: ignore
        "query": [default(""), unicode_safe],  # type: ignore
    }


@validator_args
def bulk_update_entity(
    not_empty: types.Validator,
    unicode_safe: types.Validator,
    one_of: types.Validator,
) -> types.Schema:
    entity_types = [v.entity_type for v in get_entity_managers().values()]
    actions = [opt["value"] for opt in tk.h.bulk_action_options()]

    return {
        "entity_type": [not_empty, unicode_safe, one_of(entity_types)],  # type: ignore
        "action": [not_empty, unicode_safe, one_of(actions)],  # type: ignore
        "entity_id": [not_empty, unicode_safe],
        "update_on": {
            "field": [not_empty, unicode_safe],
            "value": [not_empty, unicode_safe],
        },
    }



from _typeshed import Incomplete
from collections.abc import Generator

STDOUT: int
STDERR: int

class SocketError(Exception): ...

NPIPE_ENDED: int

def read(socket, n: int = 4096): ...
def read_exactly(socket, n): ...
def next_frame_header(socket): ...
def frames_iter(socket, tty): ...
def frames_iter_no_tty(socket) -> Generator[Incomplete, None, None]: ...
def frames_iter_tty(socket) -> Generator[Incomplete, None, None]: ...
def consume_socket_output(frames, demux: bool = False): ...
def demux_adaptor(stream_id, data): ...



from _typeshed import FileDescriptorOrPath, ReadableBuffer
from collections.abc import Callable
from typing import IO

from paramiko.message import Message
from paramiko.pkey import PKey

class DSSKey(PKey):
    p: int | None
    q: int | None
    g: int | None
    y: int | None
    x: int | None
    public_blob: None
    size: int
    def __init__(
        self,
        msg: Message | None = None,
        data: ReadableBuffer | None = None,
        filename: FileDescriptorOrPath | None = None,
        password: str | None = None,
        vals: tuple[int, int, int, int] | None = None,
        file_obj: IO[str] | None = None,
    ) -> None: ...
    def asbytes(self) -> bytes: ...
    def __hash__(self) -> int: ...
    def get_name(self) -> str: ...
    def get_bits(self) -> int: ...
    def can_sign(self) -> bool: ...
    def sign_ssh_data(self, data: bytes, algorithm: str | None = None) -> Message: ...
    def verify_ssh_sig(self, data: bytes, msg: Message) -> bool: ...
    def write_private_key_file(self, filename: FileDescriptorOrPath, password: str | None = None) -> None: ...
    def write_private_key(self, file_obj: IO[str], password: str | None = None) -> None: ...
    @staticmethod
    def generate(bits: int = 1024, progress_func: Callable[..., object] | None = None) -> DSSKey: ...



from __future__ import annotations

import json
from abc import abstractmethod
from typing import Any, TypedDict, cast

from sqlalchemy import and_, or_
from sqlalchemy.orm.attributes import InstrumentedAttribute

import ckan.plugins.toolkit as tk
from ckan import model, types
from ckan.lib.dictization import model_dictize
from ckan.lib.redis import connect_to_redis

from ckanext.bulk import const


class FilterItem(TypedDict):
    field: str
    operator: str
    value: str


class CombinedFilter(TypedDict):
    field: str
    operator: str
    values: list[str]


class FieldItem(TypedDict):
    value: str
    text: str


class UpdateItem(TypedDict):
    field: str
    value: str


class EntityMissingError(Exception):
    def __init__(self, entity_type: str):
        super().__init__(f"EntityManager type {entity_type} not found")


class EntityManager:
    entity_type = ""
    show_action = ""
    patch_action = ""
    delete_action = ""

    @classmethod
    @abstractmethod
    def get_fields(cls) -> list[FieldItem]:
        pass

    @classmethod
    @abstractmethod
    def search_entities_by_filters(
        cls, filters: list[FilterItem], global_operator: str = const.GLOBAL_AND
    ) -> list[dict[str, Any]]:
        pass

    @classmethod
    def get_entity_by_id(cls, entity_id: str) -> dict[str, Any] | None:
        try:
            group = tk.get_action(cls.show_action)(
                {"ignore_auth": True}, {"id": entity_id}
            )
        except tk.ObjectNotFound:
            return None

        return group

    @classmethod
    def combine_filters(cls, filters: list[FilterItem]) -> list[CombinedFilter]:
        combined_filters: list[CombinedFilter] = []

        current_field = ""
        current_operator = ""
        current_values = []

        for filter_item in filters:
            if (
                filter_item["field"] == current_field
                and filter_item["operator"] == current_operator
            ):
                current_values.append(filter_item["value"])
            else:
                if current_field and current_operator:
                    combined_filters.append(
                        CombinedFilter(
                            field=current_field,
                            operator=current_operator,
                            values=current_values,
                        )
                    )

                current_field = filter_item["field"]
                current_operator = filter_item["operator"]
                current_values = [filter_item["value"]]

        if current_field and current_operator:
            combined_filters.append(
                CombinedFilter(
                    field=current_field,
                    operator=current_operator,
                    values=current_values,
                )
            )

        return combined_filters

    @classmethod
    def update_entity(
        cls, entity_id: str, update_items: list[UpdateItem]
    ) -> dict[str, Any]:
        entity = cls.get_entity_by_id(entity_id)

        if not entity:
            raise tk.ObjectNotFound(f"Entity <{entity_id}> not found")

        return tk.get_action(cls.patch_action)(
            {"ignore_auth": True},
            {
                "id": entity_id,
                **cls.update_items_to_dict(update_items),
            },
        )

    @classmethod
    def update_items_to_dict(cls, update_items: list[UpdateItem]) -> dict[str, Any]:
        return {item["field"]: item["value"] for item in update_items}

    @classmethod
    def delete_entity(cls, entity_id: str) -> bool:
        tk.get_action(cls.delete_action)({"ignore_auth": True}, {"id": entity_id})

        return True

    @classmethod
    def cache_fields_to_redis(cls, fields: list[FieldItem], ttl: int = 3600):
        conn = connect_to_redis()
        conn.set(f"ckanext-bulk:fields:{cls.entity_type}", json.dumps(fields), ex=ttl)

    @classmethod
    def get_fields_from_redis(cls) -> list[FieldItem]:
        conn = connect_to_redis()

        fields = conn.get(f"ckanext-bulk:fields:{cls.entity_type}")

        if not fields:
            return []

        return json.loads(fields)


class DatasetEntityManager(EntityManager):
    entity_type = "dataset"
    show_action = "package_show"
    patch_action = "package_patch"
    delete_action = "package_delete"

    @classmethod
    def get_fields(cls) -> list[FieldItem]:
        if fields := cls.get_fields_from_redis():
            return fields

        result = tk.get_action("package_search")(
            {"ignore_auth": True},
            {"rows": 1, "include_private": True, "q": f'type:"{cls.entity_type}"'},
        )

        if not result["results"]:
            return []

        fields = [FieldItem(value=field, text=field) for field in result["results"][0]]

        cls.cache_fields_to_redis(fields)

        return fields

    @classmethod
    def search_entities_by_filters(
        cls, filters: list[FilterItem], global_operator: str = const.GLOBAL_AND
    ) -> list[dict[str, Any]]:
        """Search entities by the provided filters.

        Example of filters:
        [
            {'field': 'author', 'operator': 'is', 'value': 'Alex'},
            {'field': 'author', 'operator': 'is_not', 'value': 'John'},
            {'field': 'title', 'operator': 'contains', 'value': 'data'},
        ]

        The filters are combined with an AND operator.
        """
        q_list = []

        for f in filters:
            operator = f["operator"]

            if operator == const.OP_IS:
                q_list.append(f"{f['field']}:\"{f['value']}\"")
            elif operator == const.OP_IS_NOT:
                q_list.append(f"-{f['field']}:\"{f['value']}\"")
            elif operator == const.OP_CONTAINS:
                q_list.append(f"{f['field']}:*{f['value']}*")
            elif operator == const.OP_DOES_NOT_CONTAIN:
                q_list.append(f"-{f['field']}:*{f['value']}*")
            elif operator == const.OP_STARTS_WITH:
                q_list.append(f"{f['field']}:{f['value']}*")
            elif operator == const.OP_ENDS_WITH:
                q_list.append(f"{f['field']}:*{f['value']}")
            elif operator == const.OP_IS_EMPTY:
                q_list.append(f"-{f['field']}:[* TO *]")
            elif operator == const.OP_IS_NOT_EMPTY:
                q_list.append(f"{f['field']}:[* TO *]")

        query = f" {global_operator} ".join(q_list)
        query = f'type:"{cls.entity_type}" AND {query}'

        start = 0
        rows = 1000

        results = []

        while True:
            result = tk.get_action("package_search")(
                {"ignore_auth": True},
                {
                    "q": query,
                    "rows": rows,
                    "start": start,
                    "include_private": True,
                    "include_drafts": True,
                },
            )

            results.extend(result["results"])

            start += len(result["results"])

            if start >= result["count"]:
                break

        return results


class DatasetResourceEntityManager(EntityManager):
    entity_type = "dataset_resource"

    show_action = "resource_show"
    patch_action = "resource_patch"
    delete_action = "resource_delete"

    @classmethod
    def get_fields(cls) -> list[FieldItem]:
        if fields := cls.get_fields_from_redis():
            return fields

        fields = [
            FieldItem(value=field, text=field)
            for field in model.Resource.get_columns(extra_columns=True)
        ]

        cls.cache_fields_to_redis(fields)

        return fields

    @classmethod
    def search_entities_by_filters(
        cls, filters: list[FilterItem], global_operator: str = const.GLOBAL_AND
    ) -> list[dict[str, Any]]:
        """Search for entities by the provided filters.

        Since we are using CKAN resource_search action, we can't support
        all the operators that we have in the frontend.

        TODO: We should add support for more operators in the future.
        """
        supported_operators = [const.OP_IS]

        for f in filters:
            operator = f["operator"]

            if operator not in supported_operators:
                raise ValueError(f"Operator {operator} not supported")

        return cls.resource_search(filters, global_operator)

    @classmethod
    def resource_search(
        cls, filters: list[FilterItem], global_operator: str
    ) -> list[dict[str, Any]]:
        """This is a custom version of the core resource_search action.

        Some things we don't need were removed (hash search, term escaping),
        and some things were added (exact search for core fields, working with
        FitlerItem's).

        Note: Due to a Resource's extra fields being stored as a json blob, the
        match is made against the json string representation.  As such, false
        positives may occur:

        If the search criteria is: ::

            query = "field1:term1"

        Then a json blob with the string representation of: ::

            {"field1": "foo", "field2": "term1"}

        will match the search criteria!  This is a known short-coming of this
        approach.
        """
        context: types.Context = cast(
            types.Context,
            {"model": model, "session": model.Session, "user": tk.current_user},
        )

        base_query = model.Session.query(model.Resource).join(model.Package)
        queries = []

        for filter_item in filters:
            if not filter_item["value"]:
                continue

            field = filter_item["field"]
            value = filter_item["value"]

            # TODO: do we need escaping? The interface is available only for
            # sysadmins, so it should be safe.
            # value = misc.escape_sql_like_special_characters(value)

            if field in model.Resource.get_columns():
                queries.append(getattr(model.Resource, field) == value)

            # Resource extras are stored in a json blob.  So searching for
            # matching fields is a bit trickier.
            else:
                model_attr = cast(InstrumentedAttribute, model.Resource.extras)

                queries.append(
                    or_(
                        model_attr.ilike("""%%"%s": "%%%s%%",%%""" % (field, value)),
                        model_attr.ilike("""%%"%s": "%%%s%%"}""" % (field, value)),
                    )
                )

        if not queries:
            return []

        op_func = and_ if global_operator == const.GLOBAL_AND else or_
        base_query = base_query.filter(op_func(*queries))

        return model_dictize.resource_list_dictize(list(base_query), context)


class GroupEntityManager(EntityManager):
    entity_type = "group"

    list_action = "group_list"
    show_action = "group_show"
    patch_action = "group_patch"
    delete_action = "group_delete"

    @classmethod
    def get_fields(cls) -> list[FieldItem]:
        if fields := cls.get_fields_from_redis():
            return fields

        item_list: list[dict[str, Any]] = tk.get_action(cls.show_action)(
            {"ignore_auth": True}, {"all_fields": True, "rows": 1}
        )

        if not item_list:
            return []

        fields = [FieldItem(value=field, text=field) for field in item_list[0]]

        cls.cache_fields_to_redis(fields)

        return fields

    @classmethod
    def search_entities_by_filters(
        cls, filters: list[FilterItem], global_operator: str = const.GLOBAL_AND
    ) -> list[dict[str, Any]]:
        """Search entities by the provided filters.

        Example of filters:
        [
            {'field': 'author', 'operator': 'is', 'value': 'Alex'},
            {'field': 'author', 'operator': 'is_not', 'value': 'John'},
            {'field': 'title', 'operator': 'contains', 'value': 'data'},
        ]

        The filters are combined with an AND operator. In theory we could
        support OR operators, but we're going to keep it simple for now.

        If we need an OR operator we should use `any` instead of `all` func.
        """
        # TODO: for now we're going to fetch only 25 groups due to some
        # core restrictions.
        item_list = tk.get_action(cls.list_action)(
            {"ignore_auth": True}, {"all_fields": True}
        )

        filltered_items = []
        combined_filters = cls.combine_filters(filters)
        check_func = all if global_operator == const.GLOBAL_AND else any

        for item in item_list:
            add_item = False

            for f in combined_filters:
                operator = f["operator"]

                if f["field"] not in item:
                    break

                if operator == const.OP_IS:
                    add_item = check_func(
                        value == item[f["field"]] for value in f["values"]
                    )
                elif operator == const.OP_IS_NOT:
                    add_item = check_func(
                        value != item[f["field"]] for value in f["values"]
                    )
                elif operator == const.OP_CONTAINS:
                    add_item = check_func(
                        value in item[f["field"]] for value in f["values"]
                    )
                elif operator == const.OP_DOES_NOT_CONTAIN:
                    add_item = check_func(
                        value not in item[f["field"]] for value in f["values"]
                    )
                elif operator == const.OP_STARTS_WITH:
                    add_item = check_func(
                        item[f["field"]].startswith(value) for value in f["values"]
                    )
                elif operator == const.OP_ENDS_WITH:
                    add_item = check_func(
                        item[f["field"]].endswith(value) for value in f["values"]
                    )
                elif operator == const.OP_IS_EMPTY:
                    add_item = not item[f["field"]]
                elif operator == const.OP_IS_NOT_EMPTY:
                    add_item = bool(item[f["field"]])

                if not add_item:
                    break

            if add_item:
                filltered_items.append(item)

        return filltered_items


class OrganizationEntityManager(GroupEntityManager):
    entity_type = "organization"

    list_action = "organization_list"
    show_action = "organization_show"
    patch_action = "organization_patch"
    delete_action = "organization_delete"


def get_entity_managers() -> dict[str, type[EntityManager]]:
    return {
        "Dataset": DatasetEntityManager,
        "Resource": DatasetResourceEntityManager,
        "Organization": OrganizationEntityManager,
        "Group": GroupEntityManager,
    }


def get_entity_manager(entity_type: str) -> type[EntityManager]:
    for entity_manager in get_entity_managers().values():
        if entity_manager.entity_type == entity_type:
            return entity_manager

    raise EntityMissingError(entity_type)



from __future__ import annotations

from ckanext.bulk import const
from ckanext.bulk.entity_manager import get_entity_managers


def bulk_action_options() -> list[dict[str, str]]:
    return [
        {"value": "update", "text": "Update"},
        {"value": "delete", "text": "Delete"},
    ]


def bulk_operator_options() -> list[dict[str, str]]:
    return [
        {"text": "IS", "value": const.OP_IS},
        {"text": "IS NOT", "value": const.OP_IS_NOT},
        {"text": "CONTAINS", "value": const.OP_CONTAINS},
        {"text": "DOES NOT CONTAIN", "value": const.OP_DOES_NOT_CONTAIN},
        {"text": "STARTS WITH", "value": const.OP_STARTS_WITH},
        {"text": "ENDS WITH", "value": const.OP_ENDS_WITH},
        {"text": "IS EMPTY", "value": const.OP_IS_EMPTY},
        {"text": "IS NOT EMPTY", "value": const.OP_IS_NOT_EMPTY},
    ]


def bulk_entity_options() -> list[dict[str, str]]:
    return [
        {"value": v.entity_type, "text": k} for k, v in get_entity_managers().items()
    ]



from _typeshed import Incomplete
from typing import Any

class SwarmSpec(dict[str, Any]):
    def __init__(
        self,
        version,
        task_history_retention_limit: Incomplete | None = None,
        snapshot_interval: Incomplete | None = None,
        keep_old_snapshots: Incomplete | None = None,
        log_entries_for_slow_followers: Incomplete | None = None,
        heartbeat_tick: Incomplete | None = None,
        election_tick: Incomplete | None = None,
        dispatcher_heartbeat_period: Incomplete | None = None,
        node_cert_expiry: Incomplete | None = None,
        external_cas: Incomplete | None = None,
        name: Incomplete | None = None,
        labels: Incomplete | None = None,
        signing_ca_cert: Incomplete | None = None,
        signing_ca_key: Incomplete | None = None,
        ca_force_rotate: Incomplete | None = None,
        autolock_managers: Incomplete | None = None,
        log_driver: Incomplete | None = None,
    ) -> None: ...

class SwarmExternalCA(dict[str, Any]):
    def __init__(
        self, url, protocol: Incomplete | None = None, options: Incomplete | None = None, ca_cert: Incomplete | None = None
    ) -> None: ...



from enum import IntEnum
from typing import ClassVar, Literal

from .ImageFile import ImageFile

MAGIC: bytes

class Format(IntEnum):
    DXT1 = 0
    UNCOMPRESSED = 1

class FtexImageFile(ImageFile):
    format: ClassVar[Literal["FTEX"]]
    format_description: ClassVar[str]
    def load_seek(self, pos) -> None: ...



from __future__ import annotations

from flask import Blueprint
from flask.views import MethodView

import ckan.plugins.toolkit as tk

__all__ = ["bp"]

bp = Blueprint("bulk", __name__, url_prefix="/bulk")


@bp.errorhandler(tk.NotAuthorized)
def not_authorized_handler(_: tk.NotAuthorized) -> tuple[str, int]:
    """Generic handler for NotAuthorized exception."""
    return (
        tk.render(
            "bulk/error.html",
            {
                "code": 403,
                "content": "Not authorized to view this page",
                "name": "Not authorized",
            },
        ),
        403,
    )


def create_filter_item() -> str:
    return tk.render("bulk/snippets/filter_item.html", {"data": {}, "errors": {}})


def create_update_item() -> str:
    return tk.render("bulk/snippets/update_item.html", {"data": {}, "errors": {}})


class BulkManagerView(MethodView):
    template = "bulk/manager.html"

    def get(self):
        tk.check_access("bulk_manager", {})

        return tk.render(self.template, {"data": {}, "errors": {}})

    def post(self):
        # params = parse_params(tk.request.form)

        return tk.redirect_to("bulk.manager")


bp.add_url_rule("/manager", view_func=BulkManagerView.as_view("manager"))
bp.add_url_rule("/htmx/create_filter_item", view_func=create_filter_item)
bp.add_url_rule("/htmx/create_update_item", view_func=create_update_item)



import datetime
from _typeshed import Incomplete
from typing import Literal, NamedTuple, overload

from docker.types.daemon import CancellableStream

from .images import Image
from .resource import Collection, Model

class Container(Model):
    @property
    def name(self) -> str | None: ...
    @property
    def image(self) -> Image | None: ...
    @property
    def labels(self): ...
    @property
    def status(self) -> str: ...
    @property
    def health(self) -> str: ...
    @property
    def ports(self) -> dict[Incomplete, Incomplete]: ...
    def attach(self, **kwargs): ...
    def attach_socket(self, **kwargs): ...
    def commit(self, repository: str | None = None, tag: str | None = None, **kwargs): ...
    def diff(self): ...
    def exec_run(
        self,
        cmd,
        stdout: bool = True,
        stderr: bool = True,
        stdin: bool = False,
        tty: bool = False,
        privileged: bool = False,
        user: str = "",
        detach: bool = False,
        stream: bool = False,
        socket: bool = False,
        environment: Incomplete | None = None,
        workdir: Incomplete | None = None,
        demux: bool = False,
    ) -> ExecResult: ...
    def export(self, chunk_size: int | None = 2097152) -> str: ...
    def get_archive(
        self, path, chunk_size: int | None = 2097152, encode_stream: bool = False
    ) -> tuple[Incomplete, Incomplete]: ...
    def kill(self, signal: Incomplete | None = None): ...
    @overload
    def logs(
        self,
        *,
        stdout: bool = True,
        stderr: bool = True,
        stream: Literal[True],
        timestamps: bool = False,
        tail: Literal["all"] | int = "all",
        since: datetime.datetime | float | None = None,
        follow: bool | None = None,
        until: datetime.datetime | float | None = None,
    ) -> CancellableStream: ...
    @overload
    def logs(
        self,
        *,
        stdout: bool = True,
        stderr: bool = True,
        stream: Literal[False] = False,
        timestamps: bool = False,
        tail: Literal["all"] | int = "all",
        since: datetime.datetime | float | None = None,
        follow: bool | None = None,
        until: datetime.datetime | float | None = None,
    ) -> bytes: ...
    def pause(self) -> None: ...
    def put_archive(self, path: str, data) -> bool: ...
    def remove(self, *, v: bool = False, link: bool = False, force: bool = False) -> None: ...
    def rename(self, name: str): ...
    def resize(self, height: int, width: int): ...
    def restart(self, *, timeout: float | None = 10): ...
    def start(self) -> None: ...
    def stats(self, **kwargs): ...
    def stop(self, *, timeout: float | None = None) -> None: ...
    def top(self, *, ps_args: str | None = None) -> str: ...
    def unpause(self): ...
    def update(
        self,
        *,
        blkio_weight: int | None = None,
        cpu_period: int | None = None,
        cpu_quota: int | None = None,
        cpu_shares: int | None = None,
        cpuset_cpus: str | None = None,
        cpuset_mems: str | None = None,
        mem_limit: float | str | None = None,
        mem_reservation: float | str | None = None,
        memswap_limit: int | str | None = None,
        kernel_memory: int | str | None = None,
        restart_policy: Incomplete | None = None,
    ): ...
    def wait(self, *, timeout: float | None = None, condition: Literal["not-running", "next-exit", "removed"] | None = None): ...

class ContainerCollection(Collection[Container]):
    model: type[Container]
    @overload
    def run(
        self,
        image: str | Image,
        command: str | list[str] | None = None,
        stdout: bool = True,
        stderr: bool = False,
        remove: bool = False,
        *,
        detach: Literal[False] = False,
        **kwargs,
    ) -> bytes: ...
    @overload
    def run(
        self,
        image: str | Image,
        command: str | list[str] | None = None,
        stdout: bool = True,
        stderr: bool = False,
        remove: bool = False,
        *,
        detach: Literal[True],
        **kwargs,
    ) -> Container: ...
    def create(self, image: str, command: str | list[str] | None = None, **kwargs) -> Container: ...  # type:ignore[override]
    def get(self, container_id: str) -> Container: ...
    def list(
        self,
        all: bool = False,
        before: str | None = None,
        filters: Incomplete | None = None,
        limit: int = -1,
        since: str | None = None,
        sparse: bool = False,
        ignore_removed: bool = False,
    ): ...
    def prune(self, filters: Incomplete | None = None): ...

RUN_CREATE_KWARGS: list[str]
RUN_HOST_CONFIG_KWARGS: list[str]

class ExecResult(NamedTuple):
    exit_code: Incomplete
    output: Incomplete



from _typeshed import Incomplete, SupportsItems
from typing import Any, ClassVar, Final, Literal, overload

from pytz.tzinfo import DstTzInfo

from .caselessdict import CaselessDict
from .parser import Contentline, Contentlines
from .prop import TypesFactory

class ComponentFactory(CaselessDict[Incomplete]):
    def __init__(self, *args, **kwargs) -> None: ...

INLINE: CaselessDict[int]

class Component(CaselessDict[Incomplete]):
    name: ClassVar[str | None]
    required: ClassVar[tuple[str, ...]]
    singletons: ClassVar[tuple[str, ...]]
    multiple: ClassVar[tuple[str, ...]]
    exclusive: ClassVar[tuple[str, ...]]
    inclusive: ClassVar[tuple[tuple[str, ...], ...]]
    ignore_exceptions: ClassVar[bool]
    subcomponents: list[Incomplete]
    errors: list[str]

    def __init__(self, *args, **kwargs) -> None: ...
    def __bool__(self) -> bool: ...
    __nonzero__ = __bool__
    def is_empty(self) -> bool: ...
    @property
    def is_broken(self) -> bool: ...
    @overload
    def add(self, name: str, value: Any, *, encode: Literal[False]) -> None: ...
    @overload
    def add(self, name: str, value: Any, parameters: None, encode: Literal[False]) -> None: ...
    @overload
    def add(
        self, name: str, value: Any, parameters: SupportsItems[str, str | None] | None = None, encode: Literal[True] = True
    ) -> None: ...
    def decoded(self, name, default=[]): ...
    def get_inline(self, name, decode: bool = True): ...
    def set_inline(self, name, values, encode: bool = True) -> None: ...
    def add_component(self, component: Component) -> None: ...
    def walk(self, name: Incomplete | None = None): ...
    def property_items(self, recursive: bool = True, sorted: bool = True): ...
    @overload
    @classmethod
    def from_ical(cls, st: str, multiple: Literal[False] = False) -> Component: ...  # or any of its subclasses
    @overload
    @classmethod
    def from_ical(cls, st: str, multiple: Literal[True]) -> list[Component]: ...  # or any of its subclasses
    def content_line(self, name: str, value, sorted: bool = True) -> Contentline: ...
    def content_lines(self, sorted: bool = True) -> Contentlines: ...
    def to_ical(self, sorted: bool = True) -> bytes: ...
    def __eq__(self, other: Component) -> bool: ...  # type: ignore[override]

class Event(Component):
    name: ClassVar[Literal["VEVENT"]]

class Todo(Component):
    name: ClassVar[Literal["VTODO"]]

class Journal(Component):
    name: ClassVar[Literal["VJOURNAL"]]

class FreeBusy(Component):
    name: ClassVar[Literal["VFREEBUSY"]]

class Timezone(Component):
    name: ClassVar[Literal["VTIMEZONE"]]
    def to_tz(self) -> DstTzInfo: ...

class TimezoneStandard(Component):
    name: ClassVar[Literal["STANDARD"]]

class TimezoneDaylight(Component):
    name: ClassVar[Literal["DAYLIGHT"]]

class Alarm(Component):
    name: ClassVar[Literal["VALARM"]]

class Calendar(Component):
    name: ClassVar[Literal["VCALENDAR"]]

types_factory: Final[TypesFactory]
component_factory: Final[ComponentFactory]



from _typeshed import Incomplete

def match_tag(tag: str) -> bool: ...
def tar(
    path,
    exclude: Incomplete | None = None,
    dockerfile: Incomplete | None = None,
    fileobj: Incomplete | None = None,
    gzip: bool = False,
): ...
def exclude_paths(root, patterns, dockerfile: Incomplete | None = None): ...
def build_file_list(root): ...
def create_archive(
    root,
    files: Incomplete | None = None,
    fileobj: Incomplete | None = None,
    gzip: bool = False,
    extra_files: Incomplete | None = None,
): ...
def mkbuildcontext(dockerfile): ...
def split_path(p): ...
def normalize_slashes(p): ...
def walk(root, patterns, default: bool = True): ...

class PatternMatcher:
    patterns: Incomplete
    def __init__(self, patterns) -> None: ...
    def matches(self, filepath): ...
    def walk(self, root): ...

class Pattern:
    exclusion: bool
    dirs: Incomplete
    cleaned_pattern: Incomplete
    def __init__(self, pattern_str) -> None: ...
    @classmethod
    def normalize(cls, p): ...
    def match(self, filepath): ...



# Internal-use module for types shared by multiple modules.
# This does not match a module in docker-py.

from typing_extensions import TypeAlias

# Type alias for JSON, explained at:
# https://github.com/python/typing/issues/182#issuecomment-1320974824.
JSON: TypeAlias = dict[str, JSON] | list[JSON] | str | int | float | bool | None



from __future__ import annotations

import pytest

from ckanext.bulk.entity_manager import (
    DatasetEntityManager,
    DatasetResourceEntityManager,
    GroupEntityManager,
    OrganizationEntityManager,
)


@pytest.fixture()
def group_entity_manager():
    return GroupEntityManager


@pytest.fixture()
def organization_entity_manager():
    return OrganizationEntityManager


@pytest.fixture()
def dataset_entity_manager():
    return DatasetEntityManager


@pytest.fixture()
def dataset_resource_entity_manager():
    return DatasetResourceEntityManager



from _typeshed import Incomplete

class AddressGateway:
    gateway: Incomplete
    config: Incomplete
    def __init__(self, gateway) -> None: ...
    def create(self, params: Incomplete | None = None): ...
    def delete(self, customer_id, address_id): ...
    def find(self, customer_id, address_id): ...
    def update(self, customer_id, address_id, params: Incomplete | None = None): ...



from _typeshed import Incomplete

INDEX_NAME: str
INDEX_URL: Incomplete
TOKEN_USERNAME: str
log: Incomplete

def resolve_repository_name(repo_name): ...
def resolve_index_name(index_name): ...
def get_config_header(client, registry): ...
def split_repo_name(repo_name): ...
def get_credential_store(authconfig, registry): ...

class AuthConfig(dict[str, Incomplete]):
    def __init__(self, dct, credstore_env: Incomplete | None = None) -> None: ...
    @classmethod
    def parse_auth(cls, entries, raise_on_error: bool = False): ...
    @classmethod
    def load_config(cls, config_path, config_dict, credstore_env: Incomplete | None = None): ...
    @property
    def auths(self): ...
    @property
    def creds_store(self): ...
    @property
    def cred_helpers(self): ...
    @property
    def is_empty(self): ...
    def resolve_authconfig(self, registry: Incomplete | None = None): ...
    def get_credential_store(self, registry): ...
    def get_all_credentials(self): ...
    def add_auth(self, reg, data) -> None: ...

def resolve_authconfig(authconfig, registry: Incomplete | None = None, credstore_env: Incomplete | None = None): ...
def convert_to_hostname(url): ...
def decode_auth(auth): ...
def encode_header(auth): ...
def parse_auth(entries, raise_on_error: bool = False): ...
def load_config(
    config_path: Incomplete | None = None, config_dict: Incomplete | None = None, credstore_env: Incomplete | None = None
): ...



from _typeshed import Incomplete

from networkx.classes.graph import Graph, _Edge, _Node
from networkx.utils.backends import _dispatch
from numpy.random import RandomState

@_dispatch
def betweenness_centrality(
    G: Graph[_Node],
    k: int | None = None,
    normalized: bool = True,
    weight: str | None = None,
    endpoints: bool = False,
    seed: int | RandomState | None = None,
) -> dict[_Node, float]: ...
@_dispatch
def edge_betweenness_centrality(
    G: Graph[_Node], k: int | None = None, normalized: bool = True, weight: str | None = None, seed: Incomplete | None = None
) -> dict[_Edge[_Node], float]: ...



from __future__ import annotations

import pytest

import ckan.model as model
import ckan.plugins.toolkit as tk

from ckanext.bulk import const


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestGroupEntityManagerSearch:
    def test_filter_is(self, group_entity_manager, group_factory):
        group_factory(name="test")

        result = group_entity_manager.search_entities_by_filters(
            [{"field": "name", "operator": const.OP_IS, "value": "test"}]
        )

        assert result

    def test_filter_is_not(self, group_entity_manager, group_factory):
        group_factory(name="test")

        result = group_entity_manager.search_entities_by_filters(
            [{"field": "name", "operator": const.OP_IS_NOT, "value": "test"}]
        )

        assert not result

    def test_filter_ends_with(self, group_entity_manager, group_factory):
        group_factory(name="test")

        result = group_entity_manager.search_entities_by_filters(
            [{"field": "name", "operator": const.OP_ENDS_WITH, "value": "st"}]
        )

        assert result

    def test_filter_stars_with(self, group_entity_manager, group_factory):
        group_factory(name="test")

        result = group_entity_manager.search_entities_by_filters(
            [{"field": "name", "operator": const.OP_STARTS_WITH, "value": "te"}]
        )

        assert result

    def test_filter_contains(self, group_entity_manager, group_factory):
        group_factory(name="test")

        result = group_entity_manager.search_entities_by_filters(
            [{"field": "name", "operator": const.OP_CONTAINS, "value": "es"}]
        )

        assert result

    def test_filter_doesnt_contain(self, group_entity_manager, group_factory):
        group_factory(name="test")

        result = group_entity_manager.search_entities_by_filters(
            [{"field": "name", "operator": const.OP_DOES_NOT_CONTAIN, "value": "es"}]
        )

        assert not result

    def test_filter_is_empty(self, group_entity_manager, group_factory):
        group_factory(name="test", image_url="")

        result = group_entity_manager.search_entities_by_filters(
            [{"field": "image_url", "operator": const.OP_IS_EMPTY, "value": ""}]
        )

        assert result

    def test_filter_is_not_empty(self, group_entity_manager, group_factory):
        group_factory(name="test", image_url="test")

        result = group_entity_manager.search_entities_by_filters(
            [{"field": "image_url", "operator": const.OP_IS_NOT_EMPTY, "value": ""}]
        )

        assert result

    def test_combine_filters_1(self, group_entity_manager, group_factory):
        group_factory(name="test")

        result = group_entity_manager.search_entities_by_filters(
            [
                {"field": "name", "operator": const.OP_IS, "value": "test"},
                {"field": "name", "operator": const.OP_IS_NOT, "value": "test"},
            ]
        )

        assert not result

    def test_combine_filters_2(self, group_entity_manager, group_factory):
        group_factory(name="test")

        result = group_entity_manager.search_entities_by_filters(
            [
                {"field": "name", "operator": const.OP_IS, "value": "test"},
                {"field": "name", "operator": const.OP_ENDS_WITH, "value": "st"},
            ]
        )

        assert result

    def test_combine_filters_3(self, group_entity_manager, group_factory):
        group_factory(name="test")

        result = group_entity_manager.search_entities_by_filters(
            [
                {"field": "name", "operator": const.OP_IS, "value": "test"},
                {"field": "name", "operator": const.OP_DOES_NOT_CONTAIN, "value": "es"},
            ]
        )

        assert not result

    def test_combine_filters_4(self, group_entity_manager, group_factory):
        group_factory(name="test")

        result = group_entity_manager.search_entities_by_filters(
            [
                {"field": "name", "operator": const.OP_IS, "value": "test"},
                {"field": "name", "operator": const.OP_CONTAINS, "value": "es"},
            ]
        )

        assert result

    def test_multiple_items_1(self, group_entity_manager, group_factory):
        group_factory(name="test")
        group_factory(name="test2")

        result = group_entity_manager.search_entities_by_filters(
            [
                {"field": "name", "operator": const.OP_IS, "value": "test"},
                {"field": "name", "operator": const.OP_IS, "value": "test2"},
                {"field": "name", "operator": const.OP_IS_NOT, "value": "test3"},
                {"field": "title", "operator": const.OP_IS_NOT, "value": "test title"},
            ]
        )

        assert len(result) == 0

    def test_multiple_items_2(self, group_entity_manager, group_factory):
        group_factory(name="test")
        group_factory(name="test2")

        result = group_entity_manager.search_entities_by_filters(
            [
                {"field": "name", "operator": const.OP_IS, "value": "test"},
                {"field": "name", "operator": const.OP_IS_NOT, "value": "test2"},
            ]
        )

        assert len(result) == 1

    def test_multiple_items_3(self, group_entity_manager, group_factory):
        group_factory(name="test")
        group_factory(name="test2")

        result = group_entity_manager.search_entities_by_filters(
            [
                {"field": "name", "operator": const.OP_STARTS_WITH, "value": "te"},
                {"field": "name", "operator": const.OP_ENDS_WITH, "value": "st"},
            ]
        )

        assert len(result) == 1


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestOrganizationEntityManagerSearch:
    def test_filter_is(self, organization_entity_manager, organization_factory):
        organization_factory(name="test")

        result = organization_entity_manager.search_entities_by_filters(
            [{"field": "name", "operator": const.OP_IS, "value": "test"}]
        )

        assert result

    def test_filter_is_not(self, organization_entity_manager, organization_factory):
        organization_factory(name="test")

        result = organization_entity_manager.search_entities_by_filters(
            [{"field": "name", "operator": const.OP_IS_NOT, "value": "test"}]
        )

        assert not result

    def test_filter_ends_with(self, organization_entity_manager, organization_factory):
        organization_factory(name="test")

        result = organization_entity_manager.search_entities_by_filters(
            [{"field": "name", "operator": const.OP_ENDS_WITH, "value": "st"}]
        )

        assert result

    def test_filter_stars_with(self, organization_entity_manager, organization_factory):
        organization_factory(name="test")

        result = organization_entity_manager.search_entities_by_filters(
            [{"field": "name", "operator": const.OP_STARTS_WITH, "value": "te"}]
        )

        assert result

    def test_filter_contains(self, organization_entity_manager, organization_factory):
        organization_factory(name="test")

        result = organization_entity_manager.search_entities_by_filters(
            [{"field": "name", "operator": const.OP_CONTAINS, "value": "es"}]
        )

        assert result

    def test_filter_doesnt_contain(
        self, organization_entity_manager, organization_factory
    ):
        organization_factory(name="test")

        result = organization_entity_manager.search_entities_by_filters(
            [{"field": "name", "operator": const.OP_DOES_NOT_CONTAIN, "value": "es"}]
        )

        assert not result

    def test_filter_is_empty(self, organization_entity_manager, organization_factory):
        organization_factory(name="test", image_url="")

        result = organization_entity_manager.search_entities_by_filters(
            [{"field": "image_url", "operator": const.OP_IS_EMPTY, "value": ""}]
        )

        assert result

    def test_filter_is_not_empty(
        self, organization_entity_manager, organization_factory
    ):
        organization_factory(name="test", image_url="test")

        result = organization_entity_manager.search_entities_by_filters(
            [{"field": "image_url", "operator": const.OP_IS_NOT_EMPTY, "value": ""}]
        )

        assert result


@pytest.mark.usefixtures("with_plugins", "clean_db", "clean_index")
class TestDatasetEntityManagerSearch:
    def test_filter_is(self, dataset_entity_manager, package_factory):
        package_factory(title="test")

        result = dataset_entity_manager.search_entities_by_filters(
            [{"field": "title", "operator": const.OP_IS, "value": "test"}]
        )

        assert result

    def test_filter_is_not(self, dataset_entity_manager, package_factory):
        package_factory(title="test")

        result = dataset_entity_manager.search_entities_by_filters(
            [{"field": "title", "operator": const.OP_IS_NOT, "value": "test"}]
        )

        assert not result

    def test_filter_ends_with(self, dataset_entity_manager, package_factory):
        package_factory(title="test")

        result = dataset_entity_manager.search_entities_by_filters(
            [{"field": "title", "operator": const.OP_ENDS_WITH, "value": "st"}]
        )

        assert result

    def test_filter_stars_with(self, dataset_entity_manager, package_factory):
        package_factory(title="test")

        result = dataset_entity_manager.search_entities_by_filters(
            [{"field": "title", "operator": const.OP_STARTS_WITH, "value": "te"}]
        )

        assert result

    def test_filter_contains(self, dataset_entity_manager, package_factory):
        package_factory(title="test")

        result = dataset_entity_manager.search_entities_by_filters(
            [{"field": "title", "operator": const.OP_CONTAINS, "value": "es"}]
        )

        assert result

    def test_filter_doesnt_contain(self, dataset_entity_manager, package_factory):
        package_factory(title="test")

        result = dataset_entity_manager.search_entities_by_filters(
            [{"field": "title", "operator": const.OP_DOES_NOT_CONTAIN, "value": "es"}]
        )

        assert not result

    def test_filter_is_empty(self, dataset_entity_manager, package_factory):
        package_factory(title="test", notes="")

        result = dataset_entity_manager.search_entities_by_filters(
            [{"field": "notes", "operator": const.OP_IS_EMPTY, "value": ""}]
        )

        assert result

    def test_filter_is_not_empty(self, dataset_entity_manager, package_factory):
        package_factory(title="test", notes="test")

        result = dataset_entity_manager.search_entities_by_filters(
            [{"field": "notes", "operator": const.OP_IS_NOT_EMPTY, "value": ""}]
        )

        assert result


@pytest.mark.usefixtures("with_plugins", "clean_db", "clean_index")
class TestDatasetResourceEntityManagerSearch:
    def test_filter_is(self, dataset_resource_entity_manager, resource_factory):
        resource_factory(format="test")

        result = dataset_resource_entity_manager.search_entities_by_filters(
            [{"field": "format", "operator": const.OP_IS, "value": "test"}]
        )

        assert result

    def test_filter_is_no_match(
        self, dataset_resource_entity_manager, resource_factory
    ):
        resource_factory(format="test")

        result = dataset_resource_entity_manager.search_entities_by_filters(
            [{"field": "format", "operator": const.OP_IS, "value": "no match"}]
        )

        assert not result

    def test_operator_is_not_supported(
        self, dataset_resource_entity_manager, resource_factory
    ):
        resource_factory(format="test")

        with pytest.raises(ValueError, match="Operator contains not supported"):
            dataset_resource_entity_manager.search_entities_by_filters(
                [{"field": "format", "operator": const.OP_CONTAINS, "value": "test"}]
            )

    @pytest.mark.parametrize(
        ("field_name", "value"),
        [
            ("name", "new_name"),
            ("format", "new_format"),
            ("url", "http://example.com"),
            ("description", "test"),
        ],
    )
    def test_search_by_field(
        self, field_name, value, dataset_resource_entity_manager, resource_factory
    ):
        resource_factory(**{field_name: value})

        result = dataset_resource_entity_manager.search_entities_by_filters(
            [{"field": field_name, "operator": const.OP_IS, "value": value}]
        )

        assert result

    def test_search_similar_titles(
        self, dataset_resource_entity_manager, resource_factory
    ):
        resource_factory(name="csv data")
        resource_factory(name="information csv")

        result = dataset_resource_entity_manager.search_entities_by_filters(
            [{"field": "name", "operator": const.OP_IS, "value": "csv data"}]
        )

        assert len(result) == 1

    def test_search_title_exact_match(
        self, dataset_resource_entity_manager, resource_factory
    ):
        resource_factory(name="csv data")

        result = dataset_resource_entity_manager.search_entities_by_filters(
            [{"field": "name", "operator": const.OP_IS, "value": "csv data"}]
        )

        assert len(result) == 1

        result = dataset_resource_entity_manager.search_entities_by_filters(
            [{"field": "name", "operator": const.OP_IS, "value": "csv"}]
        )

        assert not result

    def test_search_by_extra_field(
        self, dataset_resource_entity_manager, resource_factory
    ):
        resource_factory(attribution="XXX111")

        result = dataset_resource_entity_manager.search_entities_by_filters(
            [{"field": "attribution", "operator": const.OP_IS, "value": "XXX111"}]
        )

        assert result

    def test_search_with_or_operator(
        self, dataset_resource_entity_manager, resource_factory
    ):
        resource_factory(format="CSV")
        resource_factory(format="XLSX")

        result = dataset_resource_entity_manager.search_entities_by_filters(
            [
                {"field": "format", "operator": const.OP_IS, "value": "CSV"},
                {"field": "format", "operator": const.OP_IS, "value": "XLSX"},
            ],
            const.GLOBAL_OR,
        )

        assert result


@pytest.mark.usefixtures("with_plugins", "clean_db", "clean_index")
class TestDatasetEntityManagerUpdate:
    def test_update_dataset(self, dataset_entity_manager, package_factory):
        dataset = package_factory(title="test")

        result = dataset_entity_manager.update_entity(
            dataset["id"], [{"field": "title", "value": "xxx"}]
        )

        assert result["title"] == "xxx"

    def test_update_dataset_doesnt_exist(self, dataset_entity_manager, package_factory):
        package_factory()

        with pytest.raises(tk.ObjectNotFound):
            dataset_entity_manager.update_entity("no-match", {"title": "new title"})

    def test_update_dataset_invalid_field(
        self, dataset_entity_manager, package_factory
    ):
        dataset = package_factory()

        result = dataset_entity_manager.update_entity(
            dataset["id"], [{"field": "new_field", "value": "xxx"}]
        )

        assert "new_field" not in result

    def test_update_dataset_empty_field(self, dataset_entity_manager, package_factory):
        dataset = package_factory()

        result = dataset_entity_manager.update_entity(
            dataset["id"], [{"field": "title", "value": ""}]
        )

        assert result["title"] == result["name"]

    def test_update_id_field(self, dataset_entity_manager, package_factory):
        """Try to provide an id as a filter.

        The id field is not updatable, because it will be merged into
        a final payload for the patch method and replace the id we're passing
        """
        package_factory(title="test")

        with pytest.raises(tk.ObjectNotFound):
            dataset_entity_manager.update_entity(
                "no-match", [{"field": "id", "value": "new-id"}]
            )


@pytest.mark.usefixtures("with_plugins", "clean_db", "clean_index")
class TestDatasetResourceEntityManagerUpdate:
    def test_update_dataset_resource(
        self, dataset_resource_entity_manager, resource_factory
    ):
        resource = resource_factory()

        result = dataset_resource_entity_manager.update_entity(
            resource["id"], [{"field": "format", "value": "xxx"}]
        )

        assert result["format"] == "xxx"

    def test_update_dataset_resource_doesnt_exist(
        self, dataset_resource_entity_manager, resource_factory
    ):
        with pytest.raises(tk.ObjectNotFound):
            dataset_resource_entity_manager.update_entity(
                "no-match", [{"field": "format", "value": "new"}]
            )

    def test_update_dataset_resource_new_field(
        self, dataset_resource_entity_manager, resource_factory
    ):
        """For some reason CKAN tend to save new fields for resources."""
        resource = resource_factory()

        result = dataset_resource_entity_manager.update_entity(
            resource["id"], [{"field": "new_field", "value": "xxx"}]
        )

        assert "new_field" in result

    def test_update_dataset_resource_empty_field(
        self, dataset_resource_entity_manager, resource_factory
    ):
        resource = resource_factory()

        result = dataset_resource_entity_manager.update_entity(
            resource["id"], [{"field": "format", "value": ""}]
        )

        assert not result["format"]

    def test_update_id_field(self, dataset_resource_entity_manager, resource_factory):
        resource_factory()

        with pytest.raises(tk.ObjectNotFound):
            dataset_resource_entity_manager.update_entity(
                "no-match", [{"field": "id", "value": "new-id"}]
            )


@pytest.mark.usefixtures("with_plugins", "clean_db", "clean_index")
class TestGroupEntityManagerUpdate:
    def test_update_group(self, group_entity_manager, group_factory):
        group = group_factory()

        result = group_entity_manager.update_entity(
            group["id"], [{"field": "name", "value": "xxx"}]
        )

        assert result["name"] == "xxx"

    def test_update_group_doesnt_exist(self, group_entity_manager, group_factory):
        group_factory()

        with pytest.raises(tk.ObjectNotFound):
            group_entity_manager.update_entity("no-match", {"name": "new name"})

    def test_update_group_invalid_field(self, group_entity_manager, group_factory):
        group = group_factory()

        result = group_entity_manager.update_entity(
            group["id"], [{"field": "new_field", "value": "xxx"}]
        )

        assert "new_field" not in result

    def test_update_id_field(self, group_entity_manager, group_factory):
        group_factory()

        with pytest.raises(tk.ObjectNotFound):
            group_entity_manager.update_entity("no-match", {"id": "new-id"})


@pytest.mark.usefixtures("with_plugins", "clean_db", "clean_index")
class TestOrganizationEntityManagerUpdate:
    def test_update_group(self, organization_entity_manager, organization_factory):
        organization = organization_factory()

        result = organization_entity_manager.update_entity(
            organization["id"], [{"field": "name", "value": "xxx"}]
        )

        assert result["name"] == "xxx"

    def test_update_group_doesnt_exist(
        self, organization_entity_manager, organization_factory
    ):
        organization_factory()

        with pytest.raises(tk.ObjectNotFound):
            organization_entity_manager.update_entity("no-match", {"name": "new name"})

    def test_update_group_invalid_field(
        self, organization_entity_manager, organization_factory
    ):
        organization = organization_factory()

        result = organization_entity_manager.update_entity(
            organization["id"], [{"field": "new_field", "value": "xxx"}]
        )

        assert "new_field" not in result

    def test_update_id_field(self, organization_entity_manager, organization_factory):
        organization_factory()

        with pytest.raises(tk.ObjectNotFound):
            organization_entity_manager.update_entity("no-match", {"id": "new-id"})


@pytest.mark.usefixtures("with_plugins", "clean_db", "clean_index")
class TestDatasetEntityManagerDelete:
    def test_delete_dataset(self, dataset_entity_manager, package_factory):
        dataset = package_factory()

        assert dataset_entity_manager.delete_entity(dataset["id"]) is True
        assert model.Package.get(dataset["id"]).state == model.State.DELETED  # type: ignore

    def test_delete_dataset_doesnt_exist(self, dataset_entity_manager, package_factory):
        package_factory()

        with pytest.raises(tk.ObjectNotFound):
            dataset_entity_manager.delete_entity("no-match")


@pytest.mark.usefixtures("with_plugins", "clean_db", "clean_index")
class TestDatasetResourceEntityManagerDelete:
    def test_delete_dataset_resource(
        self, dataset_resource_entity_manager, resource_factory
    ):
        resource = resource_factory()

        assert dataset_resource_entity_manager.delete_entity(resource["id"]) is True
        assert model.Resource.get(resource["id"]).state == model.State.DELETED  # type: ignore

    def test_delete_dataset_resource_doesnt_exist(
        self, dataset_resource_entity_manager, resource_factory
    ):
        resource_factory()

        with pytest.raises(tk.ObjectNotFound):
            dataset_resource_entity_manager.delete_entity("no-match")


@pytest.mark.usefixtures("with_plugins", "clean_db", "clean_index")
class TestGroupEntityManagerDelete:
    def test_delete_group(self, group_entity_manager, group_factory):
        group = group_factory()

        assert group_entity_manager.delete_entity(group["id"]) is True
        assert model.Group.get(group["id"]).state == model.State.DELETED  # type: ignore

    def test_delete_group_doesnt_exist(self, group_entity_manager, group_factory):
        group_factory()

        with pytest.raises(tk.ObjectNotFound):
            group_entity_manager.delete_entity("no-match")


@pytest.mark.usefixtures("with_plugins", "clean_db", "clean_index")
class TestOrganizationEntityManagerDelete:
    def test_delete_organization(
        self, organization_entity_manager, organization_factory
    ):
        organization = organization_factory()

        assert organization_entity_manager.delete_entity(organization["id"]) is True
        assert model.Group.get(organization["id"]).state == model.State.DELETED  # type: ignore

    def test_delete_organization_doesnt_exist(
        self, organization_entity_manager, organization_factory
    ):
        organization_factory()

        with pytest.raises(tk.ObjectNotFound):
            organization_entity_manager.delete_entity("no-match")



from ..errors import TopologicalError as TopologicalError
from ..geometry import Point, Polygon

class Cell:
    x: float
    y: float
    h: float
    centroid: Point
    distance: float
    max_distance: float
    def __init__(self, x: float, y: float, h: float, polygon: Polygon) -> None: ...
    def __lt__(self, other: Cell) -> bool: ...
    def __le__(self, other: Cell) -> bool: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __gt__(self, other: Cell) -> bool: ...
    def __ge__(self, other: Cell) -> bool: ...

def polylabel(polygon: Polygon, tolerance: float = 1.0) -> Point: ...



from collections.abc import AsyncIterator, Iterable, Mapping
from typing import Any, Literal, TypedDict, TypeVar, overload

from redis.asyncio.client import Redis
from redis.asyncio.connection import (
    BaseParser,
    ConnectCallbackT,
    Connection,
    ConnectionPool,
    Encoder,
    SSLConnection,
    _ConnectionT,
    _Sentinel,
)
from redis.asyncio.retry import Retry
from redis.commands import AsyncSentinelCommands
from redis.credentials import CredentialProvider
from redis.exceptions import ConnectionError, RedisError

_RedisT = TypeVar("_RedisT", bound=Redis[Any])

class MasterNotFoundError(ConnectionError): ...
class SlaveNotFoundError(ConnectionError): ...

class SentinelManagedConnection(Connection):
    connection_pool: ConnectionPool[Any] | None
    def __init__(
        self,
        *,
        connection_pool: ConnectionPool[Any] | None,
        # **kwargs forwarded to Connection.
        host: str = "localhost",
        port: str | int = 6379,
        socket_keepalive: bool = False,
        socket_keepalive_options: Mapping[int, int | bytes] | None = None,
        socket_type: int = 0,
        db: str | int = 0,
        password: str | None = None,
        socket_timeout: float | None = None,
        socket_connect_timeout: float | None = None,
        retry_on_timeout: bool = False,
        retry_on_error: list[type[RedisError]] | _Sentinel = ...,
        encoding: str = "utf-8",
        encoding_errors: str = "strict",
        decode_responses: bool = False,
        parser_class: type[BaseParser] = ...,
        socket_read_size: int = 65536,
        health_check_interval: float = 0,
        client_name: str | None = None,
        username: str | None = None,
        retry: Retry | None = None,
        redis_connect_func: ConnectCallbackT | None = None,
        encoder_class: type[Encoder] = ...,
        credential_provider: CredentialProvider | None = None,
    ) -> None: ...
    async def connect_to(self, address: tuple[str, int]) -> None: ...
    async def connect(self) -> None: ...

class SentinelManagedSSLConnection(SentinelManagedConnection, SSLConnection): ...

class SentinelConnectionPool(ConnectionPool[_ConnectionT]):
    is_master: bool
    check_connection: bool
    service_name: str
    sentinel_manager: Sentinel
    master_address: tuple[str, int] | None
    slave_rr_counter: int | None

    def __init__(
        self,
        service_name: str,
        sentinel_manager: Sentinel,
        *,
        ssl: bool = False,
        connection_class: type[SentinelManagedConnection] = ...,
        is_master: bool = True,
        check_connection: bool = False,
        # **kwargs ultimately forwarded to construction Connection instances.
        host: str = "localhost",
        port: str | int = 6379,
        socket_keepalive: bool = False,
        socket_keepalive_options: Mapping[int, int | bytes] | None = None,
        socket_type: int = 0,
        db: str | int = 0,
        password: str | None = None,
        socket_timeout: float | None = None,
        socket_connect_timeout: float | None = None,
        retry_on_timeout: bool = False,
        retry_on_error: list[type[RedisError]] | _Sentinel = ...,
        encoding: str = "utf-8",
        encoding_errors: str = "strict",
        decode_responses: bool = False,
        parser_class: type[BaseParser] = ...,
        socket_read_size: int = 65536,
        health_check_interval: float = 0,
        client_name: str | None = None,
        username: str | None = None,
        retry: Retry | None = None,
        redis_connect_func: ConnectCallbackT | None = None,
        encoder_class: type[Encoder] = ...,
        credential_provider: CredentialProvider | None = None,
    ) -> None: ...
    async def get_master_address(self) -> tuple[str, int]: ...
    async def rotate_slaves(self) -> AsyncIterator[tuple[str, int]]: ...

_State = TypedDict(
    "_State", {"ip": str, "port": int, "is_master": bool, "is_sdown": bool, "is_odown": bool, "num-other-sentinels": int}
)

class Sentinel(AsyncSentinelCommands):
    sentinel_kwargs: Mapping[str, Any]
    sentinels: list[Redis[Any]]
    min_other_sentinels: int
    connection_kwargs: Mapping[str, Any]
    def __init__(
        self,
        sentinels: Iterable[tuple[str, int]],
        min_other_sentinels: int = 0,
        sentinel_kwargs: Mapping[str, Any] | None = None,
        **connection_kwargs: Any,
    ) -> None: ...
    async def execute_command(self, *args: Any, once: bool = False, **kwargs: Any) -> Literal[True]: ...
    def check_master_state(self, state: _State, service_name: str) -> bool: ...
    async def discover_master(self, service_name: str) -> tuple[str, int]: ...
    def filter_slaves(self, slaves: Iterable[_State]) -> list[tuple[str, int]]: ...
    async def discover_slaves(self, service_name: str) -> list[tuple[str, int]]: ...
    @overload
    def master_for(
        self,
        service_name: str,
        redis_class: type[_RedisT],
        connection_pool_class: type[SentinelConnectionPool[Any]] = ...,
        # Forwarded to the connection pool constructor.
        **kwargs: Any,
    ) -> _RedisT: ...
    @overload
    def master_for(
        self,
        service_name: str,
        *,
        connection_pool_class: type[SentinelConnectionPool[Any]] = ...,
        # Forwarded to the connection pool constructor.
        **kwargs: Any,
    ) -> Redis[Any]: ...
    @overload
    def slave_for(
        self,
        service_name: str,
        redis_class: type[_RedisT],
        connection_pool_class: type[SentinelConnectionPool[Any]] = ...,
        # Forwarded to the connection pool constructor.
        **kwargs: Any,
    ) -> _RedisT: ...
    @overload
    def slave_for(
        self,
        service_name: str,
        *,
        connection_pool_class: type[SentinelConnectionPool[Any]] = ...,
        # Forwarded to the connection pool constructor.
        **kwargs: Any,
    ) -> Redis[Any]: ...



from _typeshed import Incomplete, StrOrBytesPath
from collections.abc import Iterable, Iterator
from typing import AnyStr, ClassVar, TypeVar, overload

from ..config import PyPIRCCommand

_T = TypeVar("_T")

class upload(PyPIRCCommand):
    description: ClassVar[str]
    username: str
    password: str
    show_response: int
    sign: bool
    identity: Incomplete
    def initialize_options(self) -> None: ...
    repository: Incomplete
    realm: Incomplete
    def finalize_options(self) -> None: ...
    def run(self) -> None: ...
    def upload_file(self, command: str, pyversion: str, filename: StrOrBytesPath) -> None: ...

@overload
def make_iterable(values: None) -> list[None]: ...
@overload
def make_iterable(values: AnyStr) -> Iterator[AnyStr]: ...
@overload
def make_iterable(values: Iterable[_T]) -> Iterator[_T]: ...
@overload
def make_iterable(values: _T) -> Iterator[_T]: ...



from typing import Any, Final, TypeVar, overload

_T = TypeVar("_T")

SEQUENCE_TYPES: Final[tuple[type[Any], ...]]
DEFAULT_ENCODING: str

def from_unicode(value: str | bytes, encoding: str = "utf-8") -> bytes: ...
def to_unicode(value: str | bytes, encoding: str = "utf-8") -> str: ...
@overload
def data_encode(data: str, encoding: str = ...) -> bytes: ...  # type: ignore[misc]
@overload
def data_encode(data: dict[Any, Any], encoding: str = ...) -> dict[Any, Any]: ...
@overload
def data_encode(data: list[Any] | tuple[Any, ...], encoding: str = ...) -> list[Any]: ...  # type: ignore[misc]
@overload
def data_encode(data: _T, encoding: str = ...) -> _T: ...



from __future__ import annotations

from typing import Any

from ckan.types import Context


def bulk_manager(context: Context, data_dict: dict[str, Any]):
    """Available to sysadmins only."""
    return {"success": False}



from _typeshed import Incomplete

class Formatter:
    pattern: Incomplete
    def __init__(self, pattern) -> None: ...
    def format(self, obj): ...
    def __call__(self, x): ...

class DecimalFormatter(Formatter):
    calcPlaces: Incomplete
    places: Incomplete
    dot: Incomplete
    comma: Incomplete
    prefix: Incomplete
    suffix: Incomplete
    def __init__(
        self,
        places: int = 2,
        decimalSep: str = ".",
        thousandSep: Incomplete | None = None,
        prefix: Incomplete | None = None,
        suffix: Incomplete | None = None,
    ) -> None: ...
    def format(self, num): ...



from _typeshed import Incomplete
from collections.abc import Callable, Generator
from typing import Final, Literal, overload
from typing_extensions import TypeAlias

from qrcode.base import RSBlock as RSBlock

_MaskPattern: TypeAlias = Literal[0, 1, 2, 3, 4, 5, 6, 7]

MODE_NUMBER: Final = 1
MODE_ALPHA_NUM: Final = 2
MODE_8BIT_BYTE: Final = 4
MODE_KANJI: Final = 8

MODE_SIZE_SMALL: Incomplete
MODE_SIZE_MEDIUM: Incomplete
MODE_SIZE_LARGE: Incomplete

ALPHA_NUM: bytes
RE_ALPHA_NUM: Incomplete
NUMBER_LENGTH: Incomplete
PATTERN_POSITION_TABLE: Incomplete
G15: Incomplete
G18: Incomplete
G15_MASK: Incomplete
PAD0: int
PAD1: int
BIT_LIMIT_TABLE: Incomplete

def BCH_type_info(data): ...
def BCH_type_number(data): ...
def BCH_digit(data): ...
def pattern_position(version): ...
def mask_func(pattern: _MaskPattern) -> Callable[[int, int], bool]: ...
def mode_sizes_for_version(version): ...
def length_in_bits(mode, version): ...
def check_version(version) -> None: ...
def lost_point(modules): ...
def optimal_data_chunks(data: str | bytes, minimum: int = 4) -> Generator[QRData, None, None]: ...
def to_bytestring(data: str | bytes) -> bytes: ...
def optimal_mode(data) -> Literal[1, 2, 4]: ...

class QRData:
    mode: Literal[1, 2, 4]
    data: bytes
    @overload
    def __init__(self, data: bytes | str, mode: Literal[1, 2, 4] | None = None, check_data: Literal[True] = True) -> None: ...
    @overload
    def __init__(self, data: bytes, mode: Literal[1, 2, 4] | None = None, *, check_data: Literal[False]) -> None: ...
    @overload
    def __init__(self, data: bytes, mode: Literal[1, 2, 4] | None, check_data: Literal[False]) -> None: ...
    def __len__(self) -> int: ...
    def write(self, buffer: BitBuffer) -> None: ...

class BitBuffer:
    buffer: Incomplete
    length: int
    def __init__(self) -> None: ...
    def get(self, index): ...
    def put(self, num, length) -> None: ...
    def __len__(self) -> int: ...
    def put_bit(self, bit) -> None: ...

def create_bytes(buffer: BitBuffer, rs_blocks: list[RSBlock]): ...
def create_data(version, error_correction, data_list): ...



from _typeshed import Incomplete
from typing import NamedTuple

class URLComponents(NamedTuple):
    scheme: Incomplete
    netloc: Incomplete
    url: Incomplete
    params: Incomplete
    query: Incomplete
    fragment: Incomplete

def create_ipam_pool(*args, **kwargs) -> None: ...
def create_ipam_config(*args, **kwargs) -> None: ...
def decode_json_header(header): ...
def compare_version(v1, v2): ...
def version_lt(v1, v2): ...
def version_gte(v1, v2): ...
def convert_port_bindings(port_bindings): ...
def convert_volume_binds(binds): ...
def convert_tmpfs_mounts(tmpfs): ...
def convert_service_networks(networks): ...
def parse_repository_tag(repo_name): ...
def parse_host(addr, is_win32: bool = False, tls: bool = False): ...
def parse_devices(devices): ...
def kwargs_from_env(environment: Incomplete | None = None): ...
def convert_filters(filters): ...
def datetime_to_timestamp(dt): ...
def parse_bytes(s): ...
def normalize_links(links): ...
def parse_env_file(env_file): ...
def split_command(command): ...
def format_environment(environment): ...
def format_extra_hosts(extra_hosts, task: bool = False): ...
def create_host_config(self, *args, **kwargs) -> None: ...



from _typeshed import SupportsGetItem
from collections.abc import Generator, Iterator
from typing import Any, overload

from openpyxl import _VisibilityType
from openpyxl.cell import _CellValue
from openpyxl.cell.cell import Cell
from openpyxl.utils.cell import _RangeBoundariesTuple
from openpyxl.workbook.workbook import Workbook
from openpyxl.worksheet.worksheet import Worksheet

def read_dimension(source) -> _RangeBoundariesTuple | None: ...

class ReadOnlyWorksheet:
    cell = Worksheet.cell
    iter_rows = Worksheet.iter_rows
    # Same as Worksheet.values
    # https://github.com/python/mypy/issues/6700
    @property
    def values(self) -> Generator[tuple[_CellValue, ...], None, None]: ...
    # Same as Worksheet.rows
    # https://github.com/python/mypy/issues/6700
    @property
    def rows(self) -> Generator[tuple[Cell, ...], None, None]: ...
    # From Worksheet.__getitem__
    @overload
    def __getitem__(self, key: int) -> tuple[Cell, ...]: ...
    @overload
    def __getitem__(self, key: slice) -> tuple[Any, ...]: ...  # tuple[AnyOf[Cell, tuple[Cell, ...]]]
    @overload
    def __getitem__(self, key: str) -> Any: ...  # AnyOf[Cell, tuple[Cell, ...], tuple[tuple[Cell, ...], ...]]
    # From Worksheet.__iter__
    def __iter__(self) -> Iterator[tuple[Cell, ...]]: ...
    parent: Workbook
    title: str
    sheet_state: _VisibilityType
    def __init__(
        self, parent_workbook: Workbook, title: str, worksheet_path, shared_strings: SupportsGetItem[int, str]
    ) -> None: ...
    def calculate_dimension(self, force: bool = False): ...
    def reset_dimensions(self) -> None: ...
    @property
    def min_row(self) -> int: ...
    @property
    def max_row(self) -> int | None: ...
    @property
    def min_column(self) -> int: ...
    @property
    def max_column(self) -> int | None: ...



from __future__ import annotations

import ckan.plugins as p
import ckan.plugins.toolkit as tk
from ckan.common import CKANConfig


@tk.blanket.actions
@tk.blanket.auth_functions
@tk.blanket.blueprints
@tk.blanket.helpers
@tk.blanket.validators
class BulkPlugin(p.SingletonPlugin):
    p.implements(p.IConfigurer)

    # IConfigurer
    def update_config(self, config_: CKANConfig):
        tk.add_template_directory(config_, "templates")
        tk.add_public_directory(config_, "public")
        tk.add_resource("assets", "bulk")



import datetime
from _typeshed import Incomplete
from datetime import tzinfo
from re import Pattern
from typing import Any, ClassVar, Final
from typing_extensions import TypeAlias

from .caselessdict import CaselessDict
from .parser import Parameters

_PropType: TypeAlias = type[Any]  # any of the v* classes in this file

DURATION_REGEX: Final[Pattern[str]]
WEEKDAY_RULE: Final[Pattern[str]]
ZERO: Final[datetime.timedelta]
HOUR: Final[datetime.timedelta]
STDOFFSET: Final[datetime.timedelta]
DSTOFFSET: Final[datetime.timedelta]
DSTDIFF: Final[datetime.timedelta]

class FixedOffset(tzinfo):
    def __init__(self, offset, name) -> None: ...
    def utcoffset(self, dt): ...
    def tzname(self, dt): ...
    def dst(self, dt): ...

class LocalTimezone(tzinfo):
    def utcoffset(self, dt): ...
    def dst(self, dt): ...
    def tzname(self, dt): ...

class vBinary:
    obj: Incomplete
    params: Incomplete
    def __init__(self, obj) -> None: ...
    def to_ical(self): ...
    @staticmethod
    def from_ical(ical): ...
    def __eq__(self, other): ...

class vBoolean(int):
    BOOL_MAP: Incomplete
    params: Incomplete
    def __new__(cls, *args, **kwargs): ...
    def to_ical(self): ...
    @classmethod
    def from_ical(cls, ical): ...

class vCalAddress(str):
    params: Incomplete
    def __new__(cls, value, encoding="utf-8"): ...
    def to_ical(self): ...
    @classmethod
    def from_ical(cls, ical): ...

class vFloat(float):
    params: Incomplete
    def __new__(cls, *args, **kwargs): ...
    def to_ical(self): ...
    @classmethod
    def from_ical(cls, ical): ...

class vInt(int):
    params: Incomplete
    def __new__(cls, *args, **kwargs): ...
    def to_ical(self): ...
    @classmethod
    def from_ical(cls, ical): ...

class vDDDLists:
    params: Incomplete
    dts: Incomplete
    def __init__(self, dt_list) -> None: ...
    def to_ical(self): ...
    @staticmethod
    def from_ical(ical, timezone: Incomplete | None = None): ...
    def __eq__(self, other): ...

class vCategory:
    cats: Incomplete
    params: Incomplete
    def __init__(self, c_list) -> None: ...
    def to_ical(self): ...
    @staticmethod
    def from_ical(ical): ...
    def __eq__(self, other): ...

class TimeBase:
    def __eq__(self, other): ...
    def __hash__(self): ...

class vDDDTypes(TimeBase):
    params: Incomplete
    dt: Incomplete
    def __init__(self, dt) -> None: ...
    def to_ical(self): ...
    @classmethod
    def from_ical(cls, ical, timezone: Incomplete | None = None): ...

class vDate(TimeBase):
    dt: Incomplete
    params: Incomplete
    def __init__(self, dt) -> None: ...
    def to_ical(self): ...
    @staticmethod
    def from_ical(ical): ...

class vDatetime(TimeBase):
    dt: Incomplete
    params: Incomplete
    def __init__(self, dt) -> None: ...
    def to_ical(self): ...
    @staticmethod
    def from_ical(ical, timezone: Incomplete | None = None): ...

class vDuration(TimeBase):
    td: Incomplete
    params: Incomplete
    def __init__(self, td) -> None: ...
    def to_ical(self): ...
    @staticmethod
    def from_ical(ical): ...
    @property
    def dt(self): ...

class vPeriod(TimeBase):
    params: Incomplete
    start: Incomplete
    end: Incomplete
    by_duration: Incomplete
    duration: Incomplete
    def __init__(self, per) -> None: ...
    def overlaps(self, other): ...
    def to_ical(self): ...
    @staticmethod
    def from_ical(ical, timezone: Incomplete | None = None): ...
    @property
    def dt(self): ...

class vWeekday(str):
    week_days: Incomplete
    relative: Incomplete
    params: Incomplete
    def __new__(cls, value, encoding="utf-8"): ...
    def to_ical(self): ...
    @classmethod
    def from_ical(cls, ical): ...

class vFrequency(str):
    frequencies: Incomplete
    params: Incomplete
    def __new__(cls, value, encoding="utf-8"): ...
    def to_ical(self): ...
    @classmethod
    def from_ical(cls, ical): ...

class vRecur(CaselessDict[Incomplete]):
    frequencies: ClassVar[list[str]]
    canonical_order: ClassVar[tuple[str, ...]]
    types: ClassVar[CaselessDict[_PropType]]
    params: Parameters
    def __init__(self, *args, **kwargs) -> None: ...
    def to_ical(self): ...
    @classmethod
    def parse_type(cls, key, values): ...
    @classmethod
    def from_ical(cls, ical) -> dict[str, Incomplete]: ...

class vText(str):
    encoding: Incomplete
    params: Incomplete
    def __new__(cls, value, encoding="utf-8"): ...
    def to_ical(self): ...
    @classmethod
    def from_ical(cls, ical): ...

class vTime(TimeBase):
    dt: Incomplete
    params: Incomplete
    def __init__(self, *args) -> None: ...
    def to_ical(self): ...
    @staticmethod
    def from_ical(ical): ...

class vUri(str):
    params: Incomplete
    def __new__(cls, value, encoding="utf-8"): ...
    def to_ical(self): ...
    @classmethod
    def from_ical(cls, ical): ...

class vGeo:
    latitude: Incomplete
    longitude: Incomplete
    params: Incomplete
    def __init__(self, geo) -> None: ...
    def to_ical(self): ...
    @staticmethod
    def from_ical(ical): ...
    def __eq__(self, other): ...

class vUTCOffset:
    ignore_exceptions: bool
    td: Incomplete
    params: Incomplete
    def __init__(self, td) -> None: ...
    def to_ical(self): ...
    @classmethod
    def from_ical(cls, ical): ...
    def __eq__(self, other): ...

class vInline(str):
    params: Incomplete
    def __new__(cls, value, encoding="utf-8"): ...
    def to_ical(self): ...
    @classmethod
    def from_ical(cls, ical): ...

class TypesFactory(CaselessDict[_PropType]):
    all_types: tuple[_PropType, ...]
    def __init__(self, *args, **kwargs) -> None: ...
    types_map: CaselessDict[str]
    def for_property(self, name: str) -> _PropType: ...
    def to_ical(self, name: str, value) -> bytes: ...
    def from_ical(self, name: str, value): ...



OP_IS = "is"
OP_IS_NOT = "is_not"
OP_CONTAINS = "contains"
OP_DOES_NOT_CONTAIN = "does_not_contain"
OP_STARTS_WITH = "starts_with"
OP_ENDS_WITH = "ends_with"
OP_IS_EMPTY = "is_empty"
OP_IS_NOT_EMPTY = "is_not_empty"

GLOBAL_AND = "AND"
GLOBAL_OR = "OR"



def i8(c: bytes) -> int: ...
def o8(i: int) -> bytes: ...
def i16le(c: bytes, o: int = 0) -> int: ...
def si16le(c: bytes, o: int = 0) -> int: ...
def si16be(c: bytes, o: int = 0) -> int: ...
def i32le(c: bytes, o: int = 0) -> int: ...
def si32le(c: bytes, o: int = 0) -> int: ...
def i16be(c: bytes, o: int = 0) -> int: ...
def i32be(c: bytes, o: int = 0) -> int: ...
def o16le(i: int) -> bytes: ...
def o32le(i: int) -> bytes: ...
def o16be(i: int) -> bytes: ...
def o32be(i: int) -> bytes: ...



from _typeshed import Incomplete

class EndpointConfig(dict[str, Incomplete]):
    def __init__(
        self,
        version,
        aliases: Incomplete | None = None,
        links: Incomplete | None = None,
        ipv4_address: Incomplete | None = None,
        ipv6_address: Incomplete | None = None,
        link_local_ips: Incomplete | None = None,
        driver_opt: Incomplete | None = None,
        mac_address: Incomplete | None = None,
    ) -> None: ...

class NetworkingConfig(dict[str, Incomplete]):
    def __init__(self, endpoints_config: Incomplete | None = None) -> None: ...

class IPAMConfig(dict[str, Incomplete]):
    def __init__(
        self, driver: str = "default", pool_configs: Incomplete | None = None, options: Incomplete | None = None
    ) -> None: ...

class IPAMPool(dict[str, Incomplete]):
    def __init__(
        self,
        subnet: Incomplete | None = None,
        iprange: Incomplete | None = None,
        gateway: Incomplete | None = None,
        aux_addresses: Incomplete | None = None,
    ) -> None: ...



from __future__ import annotations

# def bulk_required(value: Any):
#     """Verify that value is not empty."""
#     if not value or value is tk.missing:
#         raise tk.Invalid("Required")

#     return value


# def bulk_complex_validator(
#     key: types.FlattenKey,
#     data: types.FlattenDataDict,
#     errors: types.FlattenErrorDict,
#     context: types.Context,
# ):
#     """Verify that value is not empty."""
#     if not data[key]:
#         errors[key].append("Required")
#         raise tk.StopOnError



def nullTransform(): ...
def translate(dx, dy): ...
def scale(sx, sy): ...
def rotate(angle): ...
def skewX(angle): ...
def skewY(angle): ...
def mmult(A, B): ...
def inverse(A): ...
def zTransformPoint(A, v): ...
def transformPoint(A, v): ...
def transformPoints(matrix, V): ...
def zTransformPoints(matrix, V): ...
